import os
import json
import base64
import asyncio
import sqlite3
from . import solved
from . import config
from . import _http
from .ui import *
from .util import *
from .constants import *
from lxml import etree, html
from os import makedirs, path, sep, system
import sys

def pick(args, silent=False):
    pid = guess_pid(args)
    if not pid:
        print("[!] Select a problem ID")
        return
    if 'force' in args:
        force = args.force
    else:
        force = False
    prob = asyncio.run(async_pick(pid, force=force, silent=silent))
    if config.conf['auto_generate']:
        generate_code(args)

def get_cached_problem(pid):
    cur = config.db.cursor()
    q = cur.execute(f'''SELECT title, info, desc, input, output, constraints, hint, lang, spoiler, samples, accepted FROM boj WHERE pid = {pid};''').fetchone()
    if not q: return []
    prob = { 'pid':pid, 'title':q[0], 'info':json.loads(q[1]), 'desc':q[2], 'input':q[3], 'output':q[4], 'constraints':q[5], 'hint':q[6], 'lang':q[7], 'spoiler':q[8], 'samples':json.loads(q[9]), 'solved':q[10] }
    return prob

def save_problem_cache(prob):
    cur = config.db.cursor()
    cur.execute('INSERT or REPLACE INTO boj (pid, title, info, desc, input, output, constraints, hint, lang, spoiler, samples, accepted) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)', (prob['pid'], prob['title'], json.dumps(prob['info']), prob['desc'], prob['input'], prob['output'], prob['constraints'], prob['hint'], prob['lang'], prob['spoiler'], json.dumps(prob['samples']), prob['solved']))
    config.db.commit()

def generate_code(args):
    pid = guess_pid(args)
    if not pid:
        print("[!] Invalid problem ID")
        return
    template_path = path.expanduser(config.conf['template'])
    if not path.isfile(template_path):
        print("[!] Template file not found")
        return
    prob_dir = prepare_problem_dir(pid)
    ext = path.splitext(template_path)[-1]
    assert ext != "", "[!] File extension not found"
    new_path = prob_dir + sep + str(pid) + ext
    if path.exists(new_path):
        print("[!] File exists:", new_path)
        return
    inf = open(template_path, 'r')
    outf = open(new_path, 'w')
    for line in inf:
        outf.write(line)
    print(GREEN('[+] Generate {}'.format(new_path)))

async def async_pick(pid, force=False, silent=False):
    url = BOJ_HOST + '/problem/' + str(pid)
    if not silent:
        print("[+] URL:", url)
    if not force:
        prob = get_cached_problem(pid)
    else:
        prob = None
    if prob:
        show_problem(prob)
        drop_testcases(prob)
        return
    await _http.open_boj()
    try:
        resp = await _http.async_get(url)
        doc = html.fromstring(resp)
        table = doc.xpath('.//table[@id="problem-info"]')[0]
        prob = {'pid':pid}
        prob['title'] = doc.xpath('.//h1/span[@id="problem_title"]')[0].text
        prob['solved'] = STATUS_NONE
        if doc.xpath('.//span[@class="problem-label problem-label-ac"]'):
            prob['solved'] = STATUS_AC
        elif doc.xpath('.//span[@class="problem-label problem-label-wa"]'):
            prob['solved'] = STATUS_WA
        prob['info'] = {k.text.strip():v.text.strip() for k, v in zip(table.xpath('.//th'), table.xpath('.//td'))}
        prob['desc'] = get_tag_text(doc.xpath('.//div[@id="problem_description"]'))
        prob['input'] = get_tag_text(doc.xpath('.//div[@id="problem_input"]'))
        prob['output'] = get_tag_text(doc.xpath('.//div[@id="problem_output"]'))
        prob['constraints'] = get_tag_text(doc.xpath('.//div[@id="problem_limit"]'))
        prob['hint'] = get_tag_text(doc.xpath('.//div[@id="problem_hint"]'))
        lang = doc.xpath('.//div[@id="problem-lang-base64"]')
        prob['lang'] = lang[0].text if lang else None
        prob['spoiler'] = "\n".join([x.text for x in doc.xpath(".//a[@class='spoiler-link']")])
        testcases = {}
        for s in doc.xpath('.//pre[@class="sampledata"]'):
            io = s.get('id').split('-')
            num = io[2]
            if not num in testcases:
                testcases[num] = {}
            if io[1] == 'input':
                testcases[num]['in'] = s.text
            elif io[1] == 'output':
                testcases[num]['out'] = s.text
        prob['samples'] = list(testcases.values())
        save_problem_cache(prob)
        if not silent:
            show_problem(prob)
        drop_testcases(prob)
    finally:
        await _http.close_boj()

def show_problem(prob):
    print(unicode_format("\n|{:>10s} |{:>15s} |{:>10s} |{:>10s} |{:>12s} |{:>12s} |", *prob['info'].keys()))
    print(unicode_format("|{:>10s} |{:>15s} |{:>10s} |{:>10s} |{:>12s} |{:>12s} |", *prob['info'].values()))
    status = ''
    if prob['solved'] == STATUS_AC:
        status = GREEN("AC")
    elif prob['solved'] == STATUS_WA:
        status = RED("WA")
    print(BWHITE("\n[{}] {} {}\n".format(prob['pid'], prob['title'], status)))
    text_width = config.conf['text_width']
    if prob['lang']:
        multi_lang = json.loads(base64.b64decode(prob['lang']))
        for lang in multi_lang:
            print("=" * text_width)
            print(BWHITE("[+] {} : {} {}".format(lang['problem_lang_tcode'], lang['title'], status)))
            desc_text = wrap_html_text(lang['description'])
            print(desc_text)
            print(GREEN("\nInput:"))
            if lang['input']:
                in_text = wrap_html_text(lang['input'])
                print(''.join(in_text))
            print(GREEN("\nOutput:"))
            if lang['output']:
                out_text = wrap_html_text(lang['output'])
                print(''.join(out_text))
            if lang['hint']:
                hint_text = wrap_html_text(lang['hint'])
                print(''.join(hint_text))
    else:
        print(text_wrap(prob['desc'], text_width))
        print(GREEN("\nInput:"))
        if prob['input']:
            print(text_wrap(prob['input'], text_width))
        print(GREEN("\nOutput:"))
        if prob['output']:
            print(text_wrap(prob['output'], text_width))
        if prob['hint']:
            print(GREEN("\nHint:"))
            print(text_wrap(prob['hint'], text_width))

    if prob['constraints']:
        print(GREEN("\nContraints:"))
        print(prob['constraints'])
    print(BWHITE("\nSamples:"))
    idx = 1
    for tc in prob['samples']:
        print(GREEN("INPUT{}:".format(idx)))
        print(tc['in'])
        print(GREEN("OUTPUT{}:".format(idx)))
        print(tc['out'])
        idx += 1

def problem_info(args):
    pid = guess_pid(args)
    if not pid:
        print("[!] problemID is empty")
        return
    print("[+] Show problem info")
    cur = config.db.cursor()
    prob = get_cached_problem(pid)
    if not prob:
        pick(args, silent=True)
        prob = get_cached_problem(pid)
    if 'solved' in prob:
        if prob['solved'] == STATUS_AC:
            print("[+] Status : Accepted")
        elif prob['solved'] == STATUS_WA:
            print("[+] Status : Wrong answer")

    level_info = ""
    if 'level' in args and args.level:
        sinfo = solved.get_cached_level(pid)
        if not sinfo:
            sinfo = solved.get_problem_level(pid)
        level_info = "(" + sinfo['level'] + ")"

    if prob['lang']:
        multi_lang = json.loads(base64.b64decode(prob['lang']))
        for lang in multi_lang:
            print("BOJ {:d} {} {}".format(prob['pid'], lang['title'], level_info))
    else:
        print("BOJ {:d} {} {}".format(prob['pid'], prob['title'], level_info))
    print(BOJ_HOST + '/problem/' + str(pid))

def drop_testcases(prob):
    prob_dir = path.expanduser(config.conf['code_dir']) + sep + str(prob['pid'])
    makedirs(prob_dir, exist_ok=True)
    idx = 1
    for tc in prob['samples']:
        in_path = "{}{}in{:d}.txt".format(prob_dir, sep, idx)
        tc_in = tc['in'] if tc['in'] else ""
        tc_out = tc['out'] if tc['out'] else ""
        open(in_path, 'w').write(tc_in)
        out_path = "{}{}ans{:d}.txt".format(prob_dir, sep, idx)
        open(out_path, 'w').write(tc_out)
        idx += 1

def view_solutions(args):
    asyncio.run(async_view_solutions(args))

async def async_view_solutions(args):
    pid = guess_pid(args)
    if not pid:
        print("{!] problemID is empty")
        return
    await _http.open_boj()
    print("[+] View solutions")
    print(BOJ_HOST + '/problem/' + str(pid))
    try:
        # TODO support language_id configuration
        lang_id = 1001 # C++
        lang_ext = '.cpp'
        url = "https://www.acmicpc.net/status?problem_id={}&user_id=&language_id={}&result_id=4".format(pid, lang_id)
        while True:
            resp = await _http.async_get(url)
            doc = html.fromstring(resp)
            tr = doc.xpath('.//table[@id="status-table"]/tbody/tr[@id]')
            for row in tr:
                td = row.xpath('.//td')
                code_url = td[6].xpath('a[@href]')
                if len(code_url) == 0: continue
                submit_id = td[0].text
                user_info = td[1].xpath('.//span[@class]')
                if user_info and user_info[0].get('class').startswith('user-'):
                    # TODO parse atcoder color (coderTextXXXX)
                    userrank = user_info[0].get('class').split('-')[1]
                    username = ''.join(user_info[0].itertext()).ljust(20)
                    username = setcolor(userrank, username)
                else:
                    username = ''.join(td[1].itertext()).ljust(20)
                probinfo = td[2]
                result = td[3]
                prog_memory = td[4].text + td[4].xpath('.//span[@class]')[0].get('class').split('-')[0]
                prog_time = td[5].text + td[5].xpath('.//span[@class]')[0].get('class').split('-')[0]
                lang_type = code_url[0].text
                code_url = BOJ_HOST + code_url[0].get('href')
                if len(td[7]):
                    code_size = td[7].text + td[7].xpath('.//span[@class]')[0].get('class').split('-')[0]
                else:
                    code_size = ""
                if len(td[8]):
                    submit_time = td[8].xpath('.//a[@title]')[0].get('title')
                else:
                    submit_time = ""
                print("{:9s} {} {:6s} {:6s} {:6s} {:5s} {}".format(
                    submit_id, username, prog_memory, prog_time, lang_type, code_size, submit_time), end='')
                choice = input(" View? [Ynq] ").lower()
                if choice in ['y', 'yes', '']:
                    cache_dir = path.expanduser(config.conf['cache_dir']) + sep + str(pid)
                    makedirs(cache_dir, exist_ok=True)
                    cache_file = cache_dir + sep + submit_id + lang_ext
                    if not path.isfile(cache_file):
                        sol = await _http.async_get(code_url)
                        sdoc = html.fromstring(sol)
                        src = sdoc.xpath('.//textarea[@name="source"]')[0].text
                        open(cache_file, 'w').write(src)
                    system('{} "{}"'.format(config.conf['pager'], cache_file))
                elif choice in ['q', 'quit']:
                    return
            url = doc.xpath(".//a[@id='next_page']")
            if not url: break
            url = BOJ_HOST + url[0].get('href')
    finally:
        await _http.close_boj()

def set_problem_status(pid, status):
    if not pid:
        return
    if status != STATUS_NONE and status != STATUS_AC and status != STATUS_WA:
        return
    cur = config.db.cursor()
    q = cur.execute(f'''UPDATE boj SET accepted = {status} WHERE pid = {pid};''')
    config.db.commit()



from os import linesep
colors = {
    'black': "\033[30m",
    'red': "\033[31m",
    'green': "\033[32m",
    'yellow': "\033[33m",
    'blue': "\033[34m",
    'magenta': "\033[35m",
    'cyan': "\033[36m",
    'white': "\033[37m",
    'orange': "\033[33m", # yellow
    'violet': "\033[35m", # magenta
    'gray': "\033[90m",
    'bright_red': "\033[91m",
    'bright_green': "\033[92m",
    'bright_yellow': "\033[93m",
    'bright_blue': "\033[94m",
    'bright_magenta': "\033[95m",
    'bright_cyan': "\033[96m",
    'bright_white': "\033[97m",
    'nocolor': "\033[0m",
    }

def setcolor(c, msg):
    if c and c in colors:
        return colors[c] + msg + colors['nocolor']
    if c == 'legendary':
        return colors['white'] + msg[:1] + colors['red'] + msg[1:] + colors['nocolor']
    else:
        return msg;

def BLUE(msg):
    return colors['blue'] + msg + colors['nocolor']

def RED(msg):
    return colors['red'] + msg + colors['nocolor']

def GREEN(msg):
    return colors['green'] + msg + colors['nocolor']

def YELLOW(msg):
    return colors['yellow'] + msg + colors['nocolor']

def MAGENTA(msg):
    return colors['magenta'] + msg + colors['nocolor']

def CYAN(msg):
    return colors['cyan'] + msg + colors['nocolor']

def WHITE(msg):
    return colors['white'] + msg + colors['nocolor']

def BBLUE(msg):
    return colors['bright_blue'] + msg + colors['nocolor']

def BRED(msg):
    return colors['bright_red'] + msg + colors['nocolor']

def BGREEN(msg):
    return colors['bright_green'] + msg + colors['nocolor']

def BYELLOW(msg):
    return colors['bright_yellow'] + msg + colors['nocolor']

def BMAGENTA(msg):
    return colors['bright_magenta'] + msg + colors['nocolor']

def BCYAN(msg):
    return colors['bright_cyan'] + msg + colors['nocolor']

def BWHITE(msg):
    return colors['bright_white'] + msg + colors['nocolor']

def GRAY(msg):
    return colors['gray'] + msg + colors['nocolor']

def redraw(msg, end=linesep):
    print('\033[A' + msg, end=end)



from . import _http
from . import config
from .ui import *
from .util import *
from time import time
from os import unlink, path
import asyncio
import subprocess

def find_input_files(_dir):
    ins = [_dir + sep + f for f in sorted(listdir(_dir)) if path.isfile(f) and path.splitext(f)[-1] == '.txt' and f.startswith('in')]
    return ins

def compile_code(args):
    proc = subprocess.run(args, capture_output=True)
    if proc.returncode != 0:
        if proc.stdout:
            print(proc.stdout.decode())
        if proc.stderr:
            print(proc.stderr.decode())
        print(RED("[!] Compile error!"))
        exit(1)

def test(args):
    pid = guess_pid(args)
    if not pid:
        print("[!] Invalid problem ID")
        return
    prob_dir = prepare_problem_dir(pid)

    if args.input:
        filename = args.input
    else:
        filename = select_source_code(pid)
    if not filename or not path.isfile(filename):
        print("[!] File not found".format(filename))
        return
    run_path, ext = path.splitext(filename)
    ext = ext.lstrip('.')
    input_files = find_input_files(prob_dir)

    cmd = [x for x in config.conf['lang'] if x['ext'] == ext]
    if not cmd:
        print("[!] Unsupported langugage, add your configuration")
        return
    cmd = cmd[0]

    if 'compile' in cmd and cmd['compile']:
        compile_args = []
        for s in cmd['compile']:
            s = s.replace("%PROB_NUM%", run_path).replace("%SOURCE%", filename)
            compile_args.append(s)
        print("[+] Compile {}".format(filename))
        compile_code(compile_args)

    exec_args = []
    for s in cmd['cmd']:
        s = s.replace("%PROB_NUM%", run_path).replace("%SOURCE%", filename)
        exec_args.append(s)

    ac = 0
    idx = 0
    for in_file in input_files:
        idx += 1
        d = path.dirname(in_file)
        f = path.basename(in_file)
        output_file = d + sep + 'ans' + f[2:]
        if not path.isfile(output_file):
            continue
        start_time = time()
        inputs = open(in_file, "rb").read()
        try:
            proc = subprocess.run(exec_args, input=inputs, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, timeout=5)
            outputs = proc.stdout
            expected_outputs = open(output_file, "rb").read()
            same = True
            end_time = time()
            report = ''
            a = outputs.decode().splitlines()
            b = expected_outputs.decode().splitlines()
            max_length = max(len(a), len(b))
            max_width = max(20, len(max(a+b, key=len))+1)
            a += [''] * (max_length - len(a))
            b += [''] * (max_length - len(b))
            for o1, o2 in zip(a, b):
                if o1.strip() == o2.strip():
                    report += o1.ljust(max_width, ' ') * 2 + '\n'
                    continue
                else:
                    same = False
                    padding = ' ' * (max_width - len(o1))
                    report += RED(o1) + padding
                    report += GREEN(o2) + '\n'
            if same:
                ac += 1
                print(GREEN("Passed #"+str(idx)), GRAY("... {:.3f}s".format(end_time-start_time)))
            else:
                print(RED("Failed #"+str(idx)), GRAY("... {:.3f}s".format(end_time-start_time)))
                print(WHITE("=======  IN #{:d} =======".format(idx)))
                print(inputs.decode())
                print(WHITE("======= OUT #{:d} =======".format(idx)))
                print(report)
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as err:
            print(RED("Failed #{}".format(idx)))
            if err.stdout: print(err.stdout.decode())
            print(GRAY(str(err)))
    total = len(input_files)
    ac_text = "[{}/{}]".format(ac, total)
    if total == 0:
        print(RED("[!] There is no testcases"))
    elif total == ac:
        print(ac_text, GREEN("Accepted"))
    else:
        print(ac_text, RED("Wrong Answer"))
    if 'compile' in cmd and cmd['compile']:
        unlink(run_path)



import subprocess
import unicodedata
from . import config
from lxml import etree, html
from os import path, getcwd, getpid, sep, makedirs, listdir, kill
import signal


def sig_handler(signum, frame):
    job = []
    print(f"Terminated by signal : {signum}")
    signal.signal(signum, signal.SIG_DFL)
    kill(getpid(), signum)


def register_signal():
    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)


def get_tag_text(tags):
    ret = ''
    for t in tags:
        ret += t.text_content().replace('\n\t', '\n ')
    return ret.strip()

def wrap_html_text(s):
    if not s:
        return ""
    text = html.fromstring(s).text_content().replace('\n\t', '\n ')
    text = text_wrap(text, config.conf['text_width'])
    return text

def guess_pid(args):
    if 'pid' in args and args.pid:
        pid = int(args.pid)
    else:
        pid = get_cwd_info()
    return pid

def get_cwd_info():
    p = path.normpath(getcwd()).split(sep)
    if len(p) >= 1 and p[-1].isnumeric():
        return int(p[-1])
    else:
        return None

def find_source_files(_dir):
    exts = [k['ext'] for k in config.conf['lang']]
    if not exts: return
    files = [_dir + sep + f for f in listdir(_dir) if path.isfile(f) and path.splitext(f)[-1].lstrip('.') in exts]
    return files

def prepare_problem_dir(pid):
    p = path.expanduser(config.conf['code_dir'] + sep + str(pid))
    makedirs(p, exist_ok=True)
    return p

def select_source_code(pid):
    prob_path = prepare_problem_dir(pid)
    files = find_source_files(prob_path)
    if len(files) == 0:
        return None
    elif len(files) >= 2:
        print("[!] There are multiple solutions")
        return None
    return files[0]

def text_wrap(string, width):
    ret = []
    wcnt = 0
    s = ''
    for c in string:
        if unicodedata.east_asian_width(c) in ['F', 'W']:
            wcnt += 2
        else:
            wcnt += 1
        if c != '\n':
            s += c
        if wcnt >= width or c == '\n':
            wcnt = 0
            ret.append(s)
            s = ''
    if s: ret.append(s)
    return '\n'.join(ret)

def unicode_format(fmt, *args):
    new_fmt = ''
    new_args = []
    idx = 0
    for f in fmt.split('{:'):
        e = f.split('}')
        assert len(e) < 3, "Single '}' encountered"
        if len(e) == 1:
            new_fmt += e[0]
            continue
        if e[0][-1] != 's':
            new_fmt += '{:' + f
            new_args.append(args[idx])
            idx += 1
            continue
        new_fmt += '{:s}' + e[1]
        r = e[0]
        if r[0] == '>':
            direction = 1
        else:
            direction = 0
        if r[0] in ['<', '>']:
            width = int(r[1:-1])
        else:
            width = int(r[0:-1])
        wcnt = 0
        for c in args[idx]:
            if unicodedata.east_asian_width(c) in ['F', 'W']:
                wcnt += 2
            else:
                wcnt += 1
        if wcnt > width:
            wcnt = width
        if direction == 0:
            new_args.append(args[idx] + ' '*(width-wcnt))
        else:
            new_args.append(' '*(width-wcnt) + args[idx])
        idx += 1
    return new_fmt.format(*new_args)



import json
import tomli
import sqlite3
import platform
from os import makedirs, path, environ
from . import __version__

nix_path = [ '/usr/bin/chromium', '/usr/bin/chromium-browser',
             '/usr/bin/google-chrome-stable', '/usr/bin/google-chrome',
             '/snap/bin/chromium', ]
win_path = [ 'C:\Program Files (x86)\Google\Chrome\Application\chrome.exe', ]
mac_path = [ '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome',
             '/Applications/Chromium.app/Contents/MacOS/Chromium',
             '/usr/bin/google-chrome',
             '/usr/bin/google-chrome-stable',
             '/usr/bin/chromium',
             '/usr/bin/chromium-browser', ]

def_browser_path = {
    'Windows': win_path,
    'Linux': nix_path,
    'OpenBSD': nix_path,
    'Darwin': mac_path,
}
def_browser_args = [ '--start-maximized', '--no-first-run', '--disable-notifications', ]
def_user_agent = 'Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36'

DEFAULT_CONFIG = {
    'cache_dir': '~/boj/cache',
    'code_dir': '~/boj/code',
    'database': '~/.boj/cache.db',
    'template': '~/.boj/template.cpp',
    'open_in_browser': False,
    'headless_login': False,
    'pager': 'less',
    'code_open': 'open', # open, close, onlyaccepted
    'tab_width': 4,
    'auto_generate': True,
    'text_width': 70,
    'state_path': '~/.boj/state.json',
    'locale': 'ko-KR',
    'browser': '',
    'browser_args': [],
#    'browser_dir': '~/.boj/browser-cache',
    'browser_cookies': '~/.boj/cookies',
    'lang': [
        {'ext': "cpp", 'cmd': ["%PROB_NUM%"], 'compile': ["g++", "-Wall", "-W", "-std=c++17", "-O2", "-o", "%PROB_NUM%", "%SOURCE%"], 'lang_id': "C++17"},
        {'ext': "py", 'cmd': ["python3", "%SOURCE%"], 'compile': [], 'lang_id': "Python 3"},
    ],
    }

lang_ids = {
    "C++17": 84,
    "C99": 0,
    "Python 3": 28,
    "C++14": 88,
    "PyPy3": 73,
    "C++11": 49,
    "C++17 (Clang)": 85,
    "C++": 1001,
    "Java": 1002,
    "Python": 1003,
    "C": 1004,
    "Rust": 1005,
}

base_dir = environ["HOME"] + "/.boj"
config_path = base_dir + "/config.toml"
conf = {}


def find_browser_path():
    plat = platform.system()
    if not plat in def_browser_path:
        return None
    paths = def_browser_path[plat]
    for p in paths:
        if path.isfile(p):
            return p
    return None


def load_config():
    global conf, db, state, state_path
    if path.isfile(config_path):
        conf = tomli.load(open(config_path, "rb"))
    for k, v in DEFAULT_CONFIG.items():
        if not k in conf:
            conf[k] = v
    for d in [base_dir, conf['cache_dir'], conf['code_dir']]:
        if not path.isdir(path.expanduser(d)):
            makedirs(path.expanduser(d))
    state_path = path.expanduser(conf['state_path'])
    if path.isfile(state_path) and path.getsize(state_path) > 0:
        with open(state_path, 'r') as f:
            state = json.load(f)
            if 'user_agent' in state:
                conf['user_agent'] = state['user_agent']
    conf['browser_args'] += def_browser_args
    if not 'user_agent' in conf:
        conf['user_agent'] = def_user_agent

    if conf['browser_cookies']:
        conf['browser_cookies'] = path.expanduser(conf['browser_cookies'])

    if conf['browser']:
        conf['browser'] = path.expanduser(conf['browser'])
    else:
        conf['browser'] = find_browser_path()
        if not conf['browser']:
            raise Exception("Failed to detect browser environment. Define the browser within the configuration file.")

#    if conf['browser_dir']:
#        conf['browser_dir'] = path.expanduser(conf['browser_dir'])
#    if not path.isdir(conf['browser_dir']):
#        makedirs(conf['browser_dir'])

    db_path = path.expanduser(conf['database'])
    if not path.isfile(db_path):
        db = sqlite3.connect(db_path)
        cur = db.cursor()
        cur.execute('''CREATE TABLE boj (pid integer primary key, title varchar, info varchar, desc varchar, input varchar, output varchar, constraints varchar, hint varchar, lang varchar, spoiler varchar, samples varchar, accepted boolean);''')
        cur.execute('''CREATE TABLE solved (pid integer primary key, title varchar, solved_count integer, avg_try float, level varchar);''')
        db.commit()
    else:
        db = sqlite3.connect(db_path)

if not conf:
    load_config()



from . import boj
from . import login
from . import solved
from . import judge
from . import submit
from . import init
from . import  __version__
from . import util
from sys import argv
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

def main():
    parser = ArgumentParser(prog=argv[0], description="Baekjoon Online Judge & solved.ac CLI tool")
    parser.add_argument('--version', help="version", action='version', version='%(prog)s ' + __version__)
    subparsers = parser.add_subparsers(title='commands', dest='command')
    commands = {}

    _login = subparsers.add_parser('login', aliases=['l'], help="Log in to baekjoon and solved.ac", allow_abbrev=True)
    _login.set_defaults(func=login.start)

    _init = subparsers.add_parser('init', aliases=['I'], help="Setup initial file", allow_abbrev=True)
    _init.set_defaults(func=init.setup)

    _pick = subparsers.add_parser('pick', aliases=['p'], help="Pick a problem", allow_abbrev=True)
    _pick.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _pick.add_argument('-f', '--force', action='store_true', help="Update without cache data")
    _pick.set_defaults(func=boj.pick)

    _solution = subparsers.add_parser('solution', aliases=['q'], help="Get problem's solutions", allow_abbrev=True)
    _solution.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _solution.set_defaults(func=boj.view_solutions)

    _test = subparsers.add_parser('test', aliases=['t'], help="Test with testcases", allow_abbrev=True)
    _test.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _test.add_argument('-i', '--input', action='store', type=str, help="Input file")
    _test.set_defaults(func=judge.test)

    _submit = subparsers.add_parser('submit', aliases=['s'], help="Submit code", allow_abbrev=True)
    _submit.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _submit.add_argument('-i', '--input', action='store', type=str, help="Input file")
    _submit.set_defaults(func=submit.submit)

    _generate = subparsers.add_parser('generate', aliases=['g'], help="Generate source code from template")
    _generate.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _generate.set_defaults(func=boj.generate_code)

    _info = subparsers.add_parser('info', aliases=['i'], help="Show problem info")
    _info.add_argument('pid', metavar='problemID', nargs='?', action='store', type=int)
    _info.add_argument('-l', '--level', action='store_true', help="Show problem level")
    _info.set_defaults(func=boj.problem_info)

    _class = subparsers.add_parser('class', aliases=['c'], help="Pick a problem of classes", allow_abbrev=True)
    _class.add_argument('level',  metavar="class level (default number=1)", action='store', type=int, help="Pick a problem from solved.ac class")
    _class.add_argument('-e', '--essential', action='store_true', help="Essential problems only")
    _class.add_argument('-a', '--all', action='store_true', help="All problems including solved")
    _class.add_argument('-l', '--list', action='store_true', help="List title of problems")
    _class.set_defaults(func=solved.pick_class)

    _random = subparsers.add_parser('random', aliases=['r'], help="Pick a random problem from solved.ac", allow_abbrev=True)
    _random.add_argument('level_start',  metavar="level_start (default number=5)", action='store', help="Start of problem level (b, s, g, p, d, r, g5, g4, ...)")
    _random.add_argument('level_end',  metavar="level_end (default number=1)", nargs='?', action='store', help="End of problem level (b, s, g, p, d, r, g3, g2, g1, ...)")
    _random.add_argument('-l', '--list', action='store_true', help="List title of problems")
    _random.add_argument('-s', '--solved_count', action='store', type=int, help="Number of people solved")
    _random.set_defaults(func=solved.pick_random)

    util.register_signal()
    args = parser.parse_args()
    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()



from . import config
from .ui import *
from http import cookiejar
from getpass import getpass
import nodriver as uc
import asyncio
import json
import time

def start(args):
    uc.loop().run_until_complete(async_login(args))
    #return asyncio.run(async_login(args))

async def async_login(args):
    print("Sign in acmicpc.net and solve.ac")
    username = input("Username: ")
    password = getpass("Password: ")
    browser = await uc.start(headless=config.conf['headless_login'],
                             cookie=cookiejar.CookieJar(),
#                             user_data_dir=config.conf['browser_dir'],
                             browser_executable_path=config.conf['browser'],
                             browser_args=config.conf['browser_args'],
                             lang=config.conf['locale'],)
    #login_url = 'https://www.acmicpc.net/login'
    #redirect_url = 'https://solved.ac/**'
    login_url = 'https://solved.ac/login'
    tab = await browser.get(login_url)
    await tab
    time.sleep(1);
    checkbox = await tab.select('div.col-md-6:nth-child(1) > label:nth-child(1) > input:nth-child(1)')
    await checkbox.click()
    username_elem = await tab.select('div.input-group:nth-child(2) > input:nth-child(2)')
    await username_elem.send_keys(username)
    passwd_elem = await tab.select('div.input-group:nth-child(3) > input:nth-child(2)')
    await passwd_elem.send_keys(password)
    time.sleep(1);
    login_btn = await tab.select('#submit_button')
    await login_btn.click()

    for i in range(5):
        try:
            await tab.select('body')
            selector = "#login_form > div:nth-child(4) > div:nth-child(2) > a"
            element = await tab.wait_for(selector, timeout=60)
            break
        except: # TimeoutError:
            pass
    await element.click()


    selector = "#__next"
    time.sleep(1)
    for i in range(5):
        try:
            await tab.wait_for(selector, timeout=1)
            break
        except:
            pass
    state = {}
    #cookies = await browser.cookies.get_all(requests_cookie_format=False)
    cookies = await tab.send(uc.cdp.network.get_all_cookies())
    cookies = [c.to_json() for c in cookies]
    ua = await tab.evaluate('navigator.userAgent')
    state['cookies'] = cookies
    state['user_agent'] = ua
    await browser.cookies.save(config.conf['browser_cookies'])

    with open(config.state_path, 'w') as f:
        json.dump(state, f)
    print(GREEN("LOGIN SUCCESSFUL"))

    for t in browser.tabs:
        await t.close()




from . import boj
from . import _http
from . import config
from .constants import *
from lxml import etree, html
from urllib import parse
import sqlite3
import asyncio
import json
import time

levels = ('b', 's', 'g', 'p', 'd', 'r')
level_nums = ('5', '4', '3', '2', '1')
level_classes = ( 'Bronze', 'Silver', 'Gold', 'Platinum', 'Diamond', 'Ruby' )
level_roman = ( 'V', 'IV', 'III', 'II', 'I' )
level_table = {
    'b' : [1, 5], 's' : [6, 10], 'g' : [11, 15], 'p' : [16, 20], 'd' : [21, 25], 'r' : [26, 30],
    'b5' : [1], 'b4' : [2], 'b3' : [3], 'b2' : [4], 'b1' : [5],
    's5' : [6], 's4' : [7], 's3' : [8], 's2' : [9], 's1' : [10],
    'g5' : [11], 'g4' : [12], 'g3' : [13], 'g2' : [14], 'g1' : [15],
    'p5' : [16], 'p4' : [17], 'p3' : [18], 'p2' : [19], 'p1' : [20],
    'd5' : [21], 'd4' : [22], 'd3' : [23], 'd2' : [24], 'd1' : [25],
    'r5' : [26], 'r4' : [27], 'r3' : [28], 'r2' : [29], 'r1' : [30],
}

def extract_problem_from_html(tr):
    td = tr.xpath('.//td')
    prob = {}
    prob['level'] = td[0].xpath('.//img[@alt]')[0].get('alt')
    prob['pid'] = int(td[0].xpath('.//span[not(@class)]')[0].text)
    prob['title'] = ''.join(td[1].xpath(".//span[@class='__Latex__']")[0].itertext()).strip()
    prob['solved_count'] = int(td[2].xpath('.//div')[0].text.replace(',',''))
    prob['avg_try'] = float(td[3].xpath('.//div')[0].text)
    return prob

def extract_problem(item):
    prob = {}
    prob['level'] = lvnum_to_string(item['level'])
    prob['title'] = item['titleKo']
    prob['pid'] = int(item['problemId'])
    prob['solved_count'] = item['acceptedUserCount']
    prob['avg_try'] = item['averageTries']
    return prob

def lvnum_to_string(lv):
    if type(lv) == type(str()):
        lv = int(lv)
    level_name = level_classes[(lv-1)//5] + " " + level_roman[(lv-1)%5]
    return level_name


def pick_random(args):
    lv_start = str(args.level_start)
    lv_end = str(args.level_end)
    if not lv_start in level_table or not lv_end in level_table:
        print("[!] Failed to parse problem level")
        return
    lv_start = level_table[lv_start][0];
    lv_end = level_table[lv_end][-1];
    if lv_start > lv_end:
        t = lv_start
        lv_start = lv_end
        lv_end = t
    lv_range = "{}..{} ({:d}..{:d})".format(args.level_start, args.level_end, lv_start, lv_end)
#    lv_start = str(args.level_start)[:2]
#    if not lv_start[0] in levels or (len(lv_start) == 2 and not lv_start[1] in level_nums):
#        print("[!] Failed to parse problem level")
#        return
#    if 'level_end' in args and args.level_end:
#        lv_end = str(args.level_end)[:2]
#        if not lv_end[0] in levels or (len(lv_end) == 2 and not lv_end[1] in level_nums):
#            print("[!] Failed to parse problem level")
#            return
#    else:
#        lv_end = lv_start[0]

    print('[+] Random pick', lv_range)
    page = 1
    solved_count = ''
    if 'solved_count' in args and args.solved_count:
        solved_count = "solvedByGte={}&".format(args.solved_count)
    #query = parse.quote_plus(query)
    #url = '{}/problems?query={}&sort=random&direction=asc&page={:d}'.format(SOLVED_HOST, query, page)
    url = ('{}/api/v3/search/problem?query=(*{:d}..{:d})&state=unsolved&sort=random&{}' +
           't={:d}&direction=asc&page={:d}').format(SOLVED_HOST, lv_start, lv_end, solved_count, int(time.time()*1000), page)
    asyncio.run(async_query_solvedac(url, args.list))

def pick_class(args):
    level = args.level
    if level < 1 or level > 10:
        print("[!] Out of class range")
        return
    print('[+] Pick class', level)
    extra_options = ''
    if not args.all:
        extra_options += ' ~@$me'
    if args.essential:
        url = '{}/problems?query=in_class_essentials:{}{}'.format(SOLVED_HOST, level, extra_options)
    else:
        url = '{}/problems?query=in_class:{}{}'.format(SOLVED_HOST, level, extra_options)
    asyncio.run(async_query_solvedac(url, args.list))

def save_solved_list(probs):
    cur = config.db.cursor()
    for prob in probs:
        cur.execute('INSERT or REPLACE INTO solved (pid, title, solved_count, avg_try, level) VALUES (?, ?, ?, ?, ?)', (prob['pid'], prob['title'], prob['solved_count'], prob['avg_try'], prob['level']))
    config.db.commit()

def get_cached_level(pid):
    cur = config.db.cursor()
    q = cur.execute(f'''SELECT pid, title, solved_count, avg_try, level FROM solved WHERE pid={pid}''').fetchone()
    if not q: return {}
    info = { 'pid':q[0], 'title':q[1], 'solved_count':q[2], 'avg_try':q[3], 'level':q[4] }
    return info

def get_problem_level(pid):
    return asyncio.run(async_get_problem_level(pid))

async def async_get_problem_level(pid):
    await _http.open_solved()
    try:
        url = f'{SOLVED_HOST}/api/v3/search/problem?query=({pid})&page=1'
        resp = await _http.async_get(url, _http.json_headers)
        page = json.loads(resp)
        items = page['items']
        probs = []
        for i in items:
            probs += [extract_problem(i)]
        save_solved_list(probs)
        return get_cached_level(pid)
    finally:
        await _http.close_solved()

async def async_query_solvedac(url, listing=False):
    await _http.open_solved()
    try:
        resp = await _http.async_get(url, _http.json_headers)
        page = json.loads(resp)
        items = page['items']
        probs = []
        for i in items:
            probs += [extract_problem(i)]
#        tr = doc.xpath(".//table[@style='min-width:800px' and @class]//tr")
#        probs = []
#        for t in tr[1:]:
#            probs += [extract_problem_from_html(t)]
        if len(probs) == 0:
            print("[!] List is empty")
            return
        save_solved_list(probs)
        if listing:
            for prob in probs:
                print("[{:5d}] {}".format(prob['pid'], prob['title']))
        else:
            await boj.async_pick(probs[0]['pid'])
    finally:
        await _http.close_solved()



from . import config
from . import _http
from .ui import *
from .util import *
from lxml import etree, html
from urllib.parse import quote_plus
import nodriver as uc
import time
import random
import asyncio
import aiohttp

ws_url = 'wss://ws-ap1.pusher.com/app/a2cb611847131e062b32?protocol=7&client=js&version=4.2.2&flash=false'


def submit(args):
    uc.loop().run_until_complete(async_submit(args))
#    return asyncio.run(async_submit(args))

async def async_submit(args):
    global pid, ext, filename
    pid = guess_pid(args)
    if not pid:
        print("[!] Invalid problem ID")
        return
    if args.input:
        filename = args.input
    else:
        filename = select_source_code(pid)

    if not filename:
        print("[!] Failed to select a file : {}".format(filename))
        return

    if not filename:
        print("[!] Failed to select a file : {}".format(filename))
        return

    ext = path.splitext(filename)[1].lstrip('.')
    lang = [x['lang_id'] for x in config.conf['lang'] if x['ext'] == ext]
    if not lang or not lang[0] in config.lang_ids:
        print("[!] Unknown file extension")
        return

    lang_id = config.lang_ids[lang[0]]
    if not path.isfile(filename):
        print("[!] File not found : {}".format(filename))
        return

    url = 'https://www.acmicpc.net/submit/' + str(pid)
    submit_form = {
        'problem_id': str(pid),
        'language': str(lang_id),
        'code_open': config.conf['code_open'],
    }
    await async_nodriver_submit(url, submit_form, pid)
    #await async_aiohttp_submit(url, submit_form, pid)


async def is_tab_opened(tab):
    try:
        isopen = await asyncio.wait_for(tab.evaluate(expression="true"), timeout=1)
        return True
    except (ConnectionRefusedError, ConnectionError, AttributeError, asyncio.TimeoutError):
        return False


async def async_nodriver_submit(url, submit_form, pid):
    global _evt
    _evt = None
    source_code = open(filename, 'r').read()
    source_code = source_code.replace('\t', ' ' * config.conf['tab_width'])
    browser = await uc.start(headless=False,
                             browser_executable_path=config.conf['browser'],
                             browser_args=config.conf['browser_args'],
                             lang=config.conf['locale'],)
    await browser.cookies.load(config.conf['browser_cookies'])
#    tab = browser.main_tab
#    tab.add_handler(uc.cdp.network.ResponseReceived, resp_handler)

    tab = await browser.get(url)
    time.sleep(1)
    await tab
#    await tab.select('#content')
#    await tab.scroll_down(30)
    while True:
        time.sleep(0.25)
        for i in range(20):
            try:
                btn = await tab.select("#submit_button", timeout = 1)
                attr = btn.attributes
                break
            except (uc.core.connection.ProtocolException, TimeoutError):
                pass
        visible = True
        for i in range(0, len(attr), 2):
            if attr[i] == 'style' and attr[i+1] == 'display: none;':
                visible = False
                break
        if visible: break

    element = await tab.select("div.CodeMirror.cm-s-default div textarea")
    source_code = source_code.replace('\n', '\r')
    await element.send_keys(source_code)
    [await tab.send(uc.cdp.input_.dispatch_key_event("rawKeyDown", windows_virtual_key_code=46)) for i in range(100)]
#    js = "(item) => { item.dispatchEvent(new KeyboardEvent('keydown', {keyCode: 13, bubbles: true})); }"
#    await element.apply(js)
    #element = await tab.select("#cf-chl-widget-n5woi") # submit, "#submit_form > div:nth-child(7) > div > div"

    await tab.sleep(5)
    await btn.click()
    await tab
    time.sleep(1)
    for i in range(10):
        try:
            await tab.wait_for('div.col-md-12:nth-child(6)', timeout = 1)
            break
        except (uc.core.connection.ProtocolException, TimeoutError):
            pass
#    elems = await tab.select_all('#status-table > tbody:nth-child(2) > tr')
#    sids = []
#    for elem in elems:
#        attr = elem.attributes
#        for i in range(0, len(attr), 2):
#            if attr[i] == 'id':
#                sid = attr[i+1].split('-')[1]
#                sids.append(sid)
#    sid = sids[0]

#    while await is_tab_opened(tab):
#        print(browser._process)
#        print(browser.connection)
#        time.sleep(0.5)
#    for t in browser.tabs:
#        if hasattr(t, 'close'):
#            await t.close()
#    browser.stop()
#    if not await is_tab_opened(tab):
#        return

    html = await tab.get_content()
    await wait_for_status(html, pid)
    await _http.close_boj()
    time.sleep(10)

    for t in browser.tabs:
        await t.close()
#    browser.stop()

#    if _evt.response.status != 200:
#        raise Exception("The submit request failed")
#    body, b64 = await tab.send(uc.cdp.network.get_response_body(_evt.request_id))


async def resp_handler(event: uc.cdp.network.ResponseReceived):
    global _evt
    #if event.response.encoded_data_length > 0:
    _evt = event


async def async_aiohttp_submit(submit_form, pid):
    print(GREEN("[+] Submit {} ({}, {})".format(filename, lang[0], lang_id)))
    await _http.open_boj()
    try:
        resp = await _http.async_get(url)
        doc = html.fromstring(resp)
        csrf = doc.xpath('.//input[@type="hidden" and @name="csrf_key"]')[0].get("value")
        form = aiohttp.FormData()
        for k, v in submit_form.items():
            form.add_field(k, v)
        form.add_field('csrf_key', csrf)
        tab_width = config.conf['tab_width']
        source_code = open(filename, 'r').read()
        source_code = source_code.replace('\t', ' ' * tab_width)
        form.add_field('source', source_code)
        resp = await _http.async_post(url, form)
        await wait_for_status(resp, pid)
    finally:
        await _http.close_boj()


async def wait_for_status(resp, pid):
    doc = html.fromstring(resp)
    alert = doc.xpath('.//div[@class="alert-body"]')[0].text
    if alert:
        print(BWHITE(alert))
        return
# #status-table > tbody:nth-child(2)
#solution-95183678
# xpath /html/body/div[2]/div[2]/div[3]/div[6]/div/table/tbody
    js = doc.xpath('.//script[@type="text/javascript" and not(@src)]')
    for lines in js:
        for l in lines.text.splitlines():
            if l.find('solution_ids') != -1:
                sids = ''.join([c for c in l if c in '0123456789,'])
                sid = sids.split(',')[0]
    print("Waiting")
    await _http.open_boj()
    await _http.websockets(ws_url, display_submit_result, pid=pid, sid=sid)


async def display_submit_result(data):
    result = data['result']
    msg = "{:8d} : ".format(data['solution_id'])
    if result == 1:
        msg += "Prepare    "
    elif result == 2:
        msg += "Ready      "
    elif result == 3:
        if 'progress' in data:
            msg += "{:3d}%     ".format(data['progress'])
        else:
            msg += "  0%     "
    elif result == 4:
        msg += "{:s}\n{:6d}KB {:6d}ms".format(GREEN("Accepted"), data['memory'], data['time'])
    elif result == 5:
        msg += RED("Wrong Output Format")
    elif result == 6:
        msg += RED("Wrong Answer")
    elif result == 7:
        msg += RED("Time Limit Exceed")
    elif result == 8:
        msg += BLUE("Memory Limit Exceed")
    elif result == 8:
        msg += BLUE("Output Limit Exceed")
    elif result == 10:
        msg += BLUE("Runtime Error")
    elif result == 11:
        msg += BLUE("Compile Error")
    if result >= 4:
        print()
    redraw(msg)



from . import config
from . import boj
from .ui import *
from .constants import *
from http.cookies import Morsel
from datetime import datetime, timezone
from os import path
import asyncio
import aiohttp
import json
import time

default_headers = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip',
    'Accept-Language': config.conf['locale'],
    'User-Agent': config.conf['user_agent'],
    }

json_headers = {
    'Accept': 'application/json, text/plain, */',
    'Accept-Encoding': 'gzip',
    'Accept-Language': config.conf['locale'],
    'User-Agent': config.conf['user_agent'],
    }

tokens = {}
boj_session = None
solved_session = None
prev_req_time = None
req_delay = 0.750
cookie_jar = None

def add_header(newhdr, headers=default_headers):
    headers.update(newhdr)
    return headers

def get(url, headers=None):
    global prev_req_time
    resp = asyncio.run(async_get(url, headers))
    prev_req_time = time.time()
    if resp:
        return resp[0]
    else:
        return None

def post(url, data, headers=None):
    global prev_req_time
    resp = asyncio.run(async_post(url, data, headers))
    prev_req_time = time.time()
    if resp:
        return resp[0]
    else:
        return None

def GET(url, headers=None):
    return {'method':async_get, 'url':url, 'headers':headers}

def POST(url, data, headers=None):
    return {'method':async_post, 'url':url, 'data':data, 'headers':headers}

def wait_req_delay():
    if not prev_req_time:
        return
    if time.time() < prev_req_time + req_delay:
        time.sleep(prev_req_time + req_delay - time.time())

async def async_get(url, headers=None):
    if headers == None: headers = default_headers
    if url.startswith(BOJ_HOST):
        session = boj_session
    elif url.startswith(SOLVED_HOST):
        session = solved_session
    else:
        print("[!] URLs must contain {} or {}".format(BOJ_HOST, SOLVED_HOST))
        return None
    result = None
    wait_req_delay()
    async with session.get(url, headers=headers) as response:
        return await response.text()

async def async_post(url, data, headers=None):
    if headers == None: headers = default_headers
    if url.startswith(BOJ_HOST):
        session = boj_session
    elif url.startswith(SOLVED_HOST):
        session = solved_session
    else:
        print("[!] URLs must contain {} or {}".format(BOJ_HOST, SOLVED_HOST))
        return None
    session = boj_session
    result = None
    wait_req_delay()
    async with session.post(url, headers=headers, data=data) as response:
        return await response.text()

def urlsopen(urls):
    return asyncio.run(async_urlsopen(urls))

async def async_urlsopen(urls):
    tasks = []
    for u in urls:
        if u['method'] == async_get:
            tasks += [async_get(u['url'], u['headers'])]
        elif u['method'] == async_post:
            tasks += [async_post(u['url'], u['data'], u['headers'])]
    return await asyncio.gather(*tasks)


async def set_cookie():
    global cookie_jar
    cookie_jar = aiohttp.CookieJar()
    if not hasattr(config, 'state') or not 'cookies' in config.state:
        print(YELLOW("Please log in first. ") + WHITE("(boj login)"))
        return
    for c in config.state['cookies']:
        morsel = Morsel()
        morsel.set(c['name'], c['value'], c['value'])
        morsel['path'] = c['path']
        morsel['domain'] = c['domain']
        if 'expires' in c:
            t = datetime.fromtimestamp(c['expires'], tz=timezone.utc)
            morsel['expires'] = t.strftime("%a, %d %b %Y %H:%M:%S GMT")
        if 'secure' in c:
            morsel['secure'] = c['secure']
        if 'httpOnly' in c:
            morsel['httponly'] = c['httpOnly']
        if 'sameSite' in c and c['sameSite']:
            morsel['samesite'] = c['sameSite']
        cookie_jar.update_cookies({morsel.key : morsel})


async def open_boj():
    global boj_session, cookie_jar
    if cookie_jar == None:
        await set_cookie()
    if boj_session == None:
        boj_session = await aiohttp.ClientSession(cookie_jar=cookie_jar).__aenter__()

async def open_solved():
    global solved_session, cookie_jar
    if cookie_jar == None:
        await set_cookie()
    if solved_session == None:
        solved_session = await aiohttp.ClientSession(cookie_jar=cookie_jar).__aenter__()

async def close_boj():
    global boj_session
    await boj_session.__aexit__(None, None, None)
    boj_session = None

async def close_solved():
    global solved_session
    await solved_session.__aexit__(None, None, None)
    solved_session = None

async def websockets(url, callback, pid=None, sid=None):
    async with boj_session.ws_connect(url) as ws:
        result = 0
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                js = json.loads(msg.data)
                if js['event'] == 'pusher:connection_established' and sid:
                    req_js = {'event':'pusher:subscribe', \
                        'data':{'channel': 'solution-'+str(sid)}}
                    await ws.send_str(json.dumps(req_js))
                elif js['event'] == 'update':
                    data = json.loads(js['data'])
                    await callback(data)
                    if data['result'] == 4:
                        boj.set_problem_status(pid, STATUS_AC)
                        break
                    elif data['result'] > 3:
                        boj.set_problem_status(pid, STATUS_WA)
                        break
                elif js['event'] == 'pusher_internal:subscription_succeeded':
                    continue
            else:
                break;



from . import boj
from . import config
from .ui import *
import os
import tomli_w

def setup(args):
    with open(config.config_path, 'wb') as f:
        tomli_w.dump(config.DEFAULT_CONFIG, f)
    print(WHITE("Initial file generated: " + config.config_path))
